/*
BADSIEVE.C
Andrew Schulman
Microsoft Systems Journal, April 1993

16-bit (QuickWin):
cl -Ox -Gs2 -Mq -Febad16.exe badsieve.c

32-bit (WINIO):
cl386 -Ox -G3 -DUSE_WINIO -Di386=1 badsieve.c
link badsieve.obj -out:bad32.exe -debug:none winio.lib libc.lib ntdll.lib \
	kernel32.lib user32.lib
*/

#include <stdlib.h>
#ifdef USE_WINIO
#include "windows.h"
#include "winio.h"
#else
#include <stdio.h>
#endif
#include <string.h>
#include <malloc.h>
#include <limits.h>
#include <time.h>

#ifdef USE_HUGE
#define HUGE	_huge
#define ALLOC	halloc
#define USE_LONG
#else
#define HUGE	/**/
#define ALLOC	calloc
#endif

#ifdef USE_LONG
typedef unsigned long UINT;
#define MAX		LONG_MAX
#define ATOI	atol
#define MASKSTR "%lu primes <= %lu\n"
#else
typedef unsigned int UINT;
#define MAX		INT_MAX
#define ATOI	atoi
#define MASKSTR "%u primes <= %u\n"
#endif

#ifndef USE_WINIO
void fail(char *s) { puts(s); exit(1); }
#endif

void clear(UINT HUGE *p, UINT size)
{
#ifdef USE_HUGE
	UINT i;
	UINT HUGE *p2;
	for (i=size, p2=p; i--; p2++)
		*p2 = 0;
#else
	// doesn't work for huge array
	memset(p, 0, size * sizeof(UINT));
#endif
}

main(int argc, char *argv[])
{
	clock_t t;
	float runtime;
	UINT HUGE *num;
	register UINT top;
	UINT primes;
	register UINT i, j;
	int iter;
	
#ifdef __WATCOMC__
	// for now
	top = 12500;
	iter = 100;
#else
	if ((top = (argc < 2) ? 10000 : ATOI(argv[1])) >= MAX)
		fail("Number too large");
	iter = (argc < 3) ? 1 : atoi(argv[2]);
#endif	
	if (! (num = (UINT HUGE *) ALLOC(top+1, sizeof(UINT))))
		fail("Insufficient memory");
	clear(num, top+1);
	
	if (iter > 1)
		printf("%u iterations\n", iter);
	
	t = clock();
	
	while (iter--)
		for (i=2, primes=0; i<=top; i++)
			if (! num[i])
			{
				primes++;
				for (j=i+i; j<=top; j+=i)
					num[j]++;
			}

	runtime = clock() - t;

	printf(MASKSTR, primes, top);
	printf("%.2f seconds\n", runtime / CLOCKS_PER_SEC);
	return 0;
}

