/****************************************************************************
Module name: FileRev.C
*****************************************************************************/

#include <windows.h>
#include <windowsx.h>
#include <string.h>   // for strrev

/////////////////////////////////////////////////////////////////////////////

#define FILENAME "FileRev.DAT"


/////////////////////////////////////////////////////////////////////////////


int APIENTRY WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
                      LPSTR lpszCmdLine, int nCmdShow) {

   HANDLE hFile, hFileMap;
   LPVOID lpvFile;
   LPSTR lpch;
   DWORD dwFileSize;

   // Copy the input file to FILEREV.DAT so that
   //  the original is not destroyed.
   if (!CopyFile(lpszCmdLine, FILENAME, FALSE)) {
      // Copy failed.
      MessageBox(NULL, "No file specified on command line",
         "FileRev", MB_OK);
      return(0);
   }

   // Open the file for reading and writing.
   hFile = CreateFile(FILENAME, GENERIC_WRITE | GENERIC_READ, 0, NULL,
      OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
   if (hFile == INVALID_HANDLE_VALUE) {
      // File open failed.
      MessageBox(NULL, "File could not be opened.", "FileRev", MB_OK);
      return(0);
   }

   // Get the size of the file.  I am assuming that the file
   // is less than 4 gigabytes long here.
   dwFileSize = GetFileSize(hFile, NULL);

   // Create the file mapping object.  The file mapping object is made
   // one byte bigger than the file size.  This is so that a
   // zero-byte can be placed at the end to terminate the string (file).
   hFileMap = CreateFileMapping(hFile, NULL, PAGE_READWRITE,
      0, dwFileSize + 1, NULL);
   if (hFileMap == NULL) {
      // File Mapping  open failed.
      MessageBox(NULL, "File map could not be opened.", "FileRev", MB_OK);
      return(0);
   }

   // Get the addess to where the first byte of the file is mapped
   // into memory.
   lpvFile = MapViewOfFile(hFileMap, FILE_MAP_WRITE, 0, 0, 0);
   if (lpvFile == NULL) {
      // Map view of file failed.
      MessageBox(NULL, "Could not map view of file.", "FileRev", MB_OK);
      return(0);
   }

   // Put a zero byte at the very end of the file.
   ((LPSTR) lpvFile)[dwFileSize] = 0;

   // Reverse the contents of the file.
   strrev(lpvFile);


   // Convert all "\n\r" combinations back to their original "\r\n".
   // to preserve the normal end-of-line sequence.
   lpch = strchr(lpvFile, '\n'); // Find first occurance of '\n'.
   while (lpch != NULL) {
      // While we have found an occurance...
      *lpch++ = '\r'; // Change the '\n' to '\r'.
      *lpch++ = '\n'; // Change the '\r' to '\n'.
      lpch = strchr(lpch, '\n'); // Find the next occurance.
   }


   // Remove the trailing zero-byte added earlier by positioning
   // the file pointer to the end of the file not including the
   // zero-byte and setting the end-of-file.
   SetFilePointer(hFile, dwFileSize, NULL, FILE_BEGIN);
   SetEndOfFile(hFile);

   // Cleanup everything before exiting.
   UnmapViewOfFile(lpvFile);
   CloseHandle(hFileMap);
   CloseHandle(hFile);

   // Spawn NOTEPAD so that we can see the fruits our labor.
   WinExec("NOTEPAD.EXE " FILENAME, SW_SHOW);
   return(0);
}


///////////////////////////////// End of file /////////////////////////////
