/*******************************************************************************
Module name: MMFShare.C
*******************************************************************************/

#include <windows.h>
#include <windowsx.h>
#include "MMFShare.H"


/////////////////////////////////////////////////////////////////////////////


BOOL Dlg_OnInitDialog (HWND hwnd, HWND hwndFocus, LPARAM lParam) {
   // Initialize the EDIT control with some test data.
   Edit_SetText(GetDlgItem(hwnd, ID_DATA), "Some test data");

   // Disable the "Close file" button because the file cannot
   // be closed if it was never created or opened.
   Button_Enable(GetDlgItem(hwnd, ID_CLOSEFILE), FALSE);
   return(TRUE);
}


/////////////////////////////////////////////////////////////////////////////


void Dlg_OnCommand (HWND hwnd, int id, HWND hwndCtl, UINT codeNotify) {
   // Handle of the open memory mapped file.
   static HANDLE s_hFileMap = NULL;

   HANDLE hFileMapT;

   switch (id) {
      case ID_CREATEFILE:
         if (codeNotify != BN_CLICKED)
            break;

         // Create an in-memory memory mapped file using the contents of
         // the EDIT control as the contents of the file.
         // The is at most 4K big and has a name of "MMFSharedData".
         s_hFileMap = CreateFileMapping((HANDLE) 0xFFFFFFFF, NULL,
                            PAGE_READWRITE, 0, 4 * 1024, "MMFSharedData");

         if (s_hFileMap != NULL) {

            if (GetLastError() == ERROR_ALREADY_EXISTS) {
               MessageBox(hwnd,
                  "Mapping already exists - not created.",
                  NULL, MB_OK);
               CloseHandle(s_hFileMap);

            } else {

               // File mapping created successfully.

               // Map a view of the file into the address space.
               LPVOID lp = MapViewOfFile(s_hFileMap,
                  FILE_MAP_READ | FILE_MAP_WRITE, 0, 0, 0);

               if (lp != NULL) {
                  // View mapped successfully, put contents of
                  // EDIT control in the the memory mapped file.
                  Edit_GetText(GetDlgItem(hwnd, ID_DATA),
                               lp, 4 * 1024);

                  // Unmap the view - this also protects the data
                  // from wayward pointers.
                  UnmapViewOfFile(lp);

                  // The user cannot create another file right now
                  Button_Enable(hwndCtl, FALSE);

                  // The user close the file.
                  Button_Enable(GetDlgItem(hwnd, ID_CLOSEFILE),
                                TRUE);

               } else {
                  MessageBox(hwnd, "Can't map view of file.",
                             NULL, MB_OK);
               }
            }

         } else {
            MessageBox(hwnd, "Can't create file mapping.",
                       NULL, MB_OK);
         }
         break;


      case ID_CLOSEFILE:
         if (codeNotify != BN_CLICKED)
            break;

         if (CloseHandle(s_hFileMap)) {
            // User closed the file.  A new file can be created
            // and the file can't be closed again.
            Button_Enable(GetDlgItem(hwnd, ID_CREATEFILE), TRUE);
            Button_Enable(hwndCtl, FALSE);
         }
         break;


      case ID_OPENFILE:
         if (codeNotify != BN_CLICKED)
            break;

         // See if a memory mapped file called
         // "MMFSharedData" already exists.
         hFileMapT = OpenFileMapping(FILE_MAP_READ | FILE_MAP_WRITE,
                                     FALSE, "MMFSharedData");

         if (hFileMapT != NULL) {
            // Memory mapped file does exists.  Map a view of it
            // into the process address space.
            LPVOID lp = MapViewOfFile(hFileMapT, FILE_MAP_READ,
                                      0, 0, 0);

            if (lp != NULL) {

               // Put the contents of the file into the EDIT control.
               Edit_SetText(GetDlgItem(hwnd, ID_DATA), lp);
               UnmapViewOfFile(lp);

            } else {
               MessageBox(hwnd, "Can't map view.", NULL, MB_OK);
            }

            CloseHandle(hFileMapT);

         } else {
            MessageBox(hwnd, "Can't open mapping.", NULL, MB_OK);
         }
         break;

      case IDCANCEL:
         EndDialog(hwnd, id);
         break;
   }
}


///////////////////////////////////////////////////////////////////////////


BOOL CALLBACK DlgProc (HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
   BOOL fProcessed = TRUE;

   switch (uMsg) {
      HANDLE_MSG(hDlg, WM_INITDIALOG, Dlg_OnInitDialog);
      HANDLE_MSG(hDlg, WM_COMMAND, Dlg_OnCommand);

      default:
         fProcessed = FALSE;
         break;
   }
   return(fProcessed);
}


///////////////////////////////////////////////////////////////////////////


int APIENTRY WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
                      LPSTR lpszCmdLine, int nCmdShow) {

   DialogBox(hInstance, MAKEINTRESOURCE(DLG_MMFSHARE), NULL, DlgProc);
   return(0);
}


///////////////////////////////// End of file /////////////////////////////
