/* 
OKSIEVE.C
Andrew Schulman
Microsoft Systems Journal, April 1993

16-bit (QuickWin):
cl -Ox -Gs2 -Mq -Feok16.exe oksieve.c

32-bit (WINIO):
cl386 -Ox -G3 -DUSE_WINIO -Di386=1 oksieve.c
link oksieve.obj -out:ok32.exe -debug:none winio.lib libc.lib ntdll.lib \
	kernel32.lib user32.lib
*/

#include <stdlib.h>
#ifdef USE_WINIO
#include "windows.h"
#include "winio.h"
#else
#include <stdio.h>
#endif
#include <string.h>
#include <math.h>
#include <malloc.h>
#include <limits.h>
#include <time.h>

#ifdef USE_HUGE
#define HUGE	_huge
#define ALLOC	halloc
#define USE_LONG
#else
#define HUGE	/**/
#define ALLOC	calloc
#endif

#ifdef USE_LONG
typedef unsigned long UINT;
#define MAX		(2L * LONG_MAX)
#define ATOI	atol
#define MASKSTR "%lu primes <= %lu\n"
#else
typedef unsigned int UINT;
#define MAX		(2L * INT_MAX)
#define ATOI	atoi
#define MASKSTR "%u primes <= %u\n"
#endif

#define N(x)				((x) >> 1)

#ifndef USE_WINIO
void fail(char *s) { puts(s); exit(1); }
#endif

void clear(UINT HUGE *p, UINT size)
{
#ifdef USE_HUGE
	UINT i;
	UINT HUGE *p2;
	for (i=size>>1, p2=p; i--; p2++)
		*p2 = 0;
#else
	// doesn't work for huge array
	memset(p, 0, size * sizeof(UINT));
#endif
}

main(int argc, char *argv[])
{
	clock_t t;
	float runtime;
	UINT HUGE *num;
	register UINT top;
	UINT primes;
	UINT sqrt_top;
	register UINT i, j;
	register UINT i2;
	int iter;

#ifdef __WATCOMC__
	// for now
	top = 12500;
	iter = 100;
#else
	if ((top = (argc < 2) ? 10000 : ATOI(argv[1])) >= MAX)
		fail("Number too large");
	iter = (argc < 3) ? 1 : atoi(argv[2]);
#endif	
	if (! (num = (UINT HUGE *) ALLOC(N(top+1), sizeof(UINT))))
		fail("Insufficient memory");
	clear(num, N(top+1));
	
	if (iter > 1)
		printf("%u iterations\n", iter);
	
	t = clock();
	
	sqrt_top = sqrt(top);
	
	while (iter--)
	{
		// mark (and sweep up to sqrt_top)
		for (i=3, primes=1; i<=sqrt_top; i += 2)
			if (! num[N(i)])
			{
				primes++;
				for (i2=i+i, j=i2+i; j<=top; j+=i2) // skip mult of 2
					num[N(j)]++;
			}
				
		// sweep from sqrt_top to top
		for ( ; i<=top; i += 2)
			if (! num[N(i)])
				primes++;
	}

	runtime = clock() - t;

	printf(MASKSTR, primes, top);
	printf("%.2f seconds\n", runtime / CLOCKS_PER_SEC);
	return 0;
}


