/*
WINSIZE.C -- Show various Windows "sizes"
Andrew Schulman
Microsoft Systems Journal, April 1993

16-bit (WINSZ16.EXE):
cl -Gs2 -Ox -GA -Fewinsz16.exe winsize.c winsize.def 

32-bit (WINSZ32.EXE):
cl386 -DWIN32 -G3 -Di386=1 -c -Fc winsize.c
link winsize.obj -debug:none -subsystem:windows -out:winsz32.exe \
	libc.lib ntdll.lib kernel32.lib user32.lib
*/

#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "windows.h"

// high bit of GetVersion() set if Win32s;
// high bit clear if Windows NT
#define WIN32S_BIT		0x80000000

static char buf[1024] = {0};
static char title[64] = {0};

#define print(str, x) { \
	char tmp[128]; \
	wsprintf(tmp, (str), (x)); \
	strcat(buf, tmp); \
	}

unsigned short getds(void)
{
	_asm xor ax, ax
	_asm mov ax, ds
}

unsigned short getcs(void)
{
	_asm xor ax, ax
	_asm mov ax, cs
}

int PASCAL WinMain(HANDLE hInstance, HANDLE hPrevInstance, 
	LPSTR lpCmdLine, int nCmdShow)
{
	HANDLE h;
	DWORD dwVers;
	print("sizeof(int) = %d\n", sizeof(int));
	print("sizeof(size_t) = %d\n", sizeof(size_t));
	print("sizeof(HANDLE) = %d\n", sizeof(HANDLE));
	print("sizeof(void near *) = %d\n", sizeof(void near *));
	print("GlobalAlloc = %X\n", h = GlobalAlloc(GMEM_MOVEABLE, 1024));
	print("GlobalLock = %08lX\n", GlobalLock(h));
	print("LocalAlloc = %X\n", h = LocalAlloc(GMEM_MOVEABLE, 1024));
	print("LocalLock = %X\n", LocalLock(h));
	print("malloc = %X\n", malloc(1024));
	print("Typical API function pointer = %08lX\n", GetVersion);
	print("DS = %Xh\n", getds());
	print("CS = %Xh\n", getcs());
	print("Code at %Xh\n", getcs);
	dwVers = GetVersion();
	if (sizeof(int) == 4) // Win32 API
	{
		if (dwVers & WIN32S_BIT)	// Win32s
		{
			wsprintf(title, "WINSIZE (%d Bits) - Windows %d.%02d", 
				sizeof(int) * CHAR_BIT,
				LOBYTE(dwVers), HIBYTE(dwVers));
		}
		else	// Windows NT	<<IS THIS RIGHT????>>
		{
			wsprintf(title, "WINSIZE (%d Bits) - Windows NT %d.%02d", 
				sizeof(int) * CHAR_BIT,
				LOBYTE(dwVers), HIBYTE(dwVers));
		}
	}
	else	// Win16 API
	{
		wsprintf(title, "WINSIZE (%d Bits) - Windows %d.%02d", 
			sizeof(int) * CHAR_BIT,
			LOBYTE(dwVers), HIBYTE(dwVers));
	}
	MessageBox(0, buf, title, MB_OK);
	return 0;
}


